/*
	this file is under Public Domain
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef struct {
	size_t length;  /* maximam length of string */ 
	size_t count;   /* current length of string */
	size_t delta;
	char *str;
} String;

int String_alloc(String *s)
{
	char *p;
	p = (char *)realloc(s->str, sizeof(char) * (s->length + 1));
	if (!p) {
		free(s->str);
		return 1;
	}
	s->str = p;
	return 0;
}

int String_init(String *s, size_t length)
{
	s->length = length;
	s->count = 0;
	s->delta = length / 2 + 1;
	s->str = NULL;
	if (String_alloc(s)) {
		return 1;
	}
	s->str[0] = '\0';
	return 0;
}

int String_append_char(String *s, char c)
{
	if (s->count >= s->length) {
		s->length += s->delta;
		if (String_alloc(s)) {
			return 1;
		}
	}
	s->str[s->count] = c;
	s->count ++;
	s->str[s->count] = '\0';
	return 0;
}

int String_append(String *s, const char *st, size_t size)
{
	size_t l = s->count + size;
	if (l > s->length) {
		s->length += ((l - s->length) / s->delta + 1) * s->delta;
		if (String_alloc(s)) {
			return 1;
		}
	}
	strcat(s->str, st);
	s->count += size;
	return 0;
}


char *sjis2jis(char *s)
{
	enum {ASCII, KANJI} seq = ASCII;
	String ost;
	unsigned char *c, h;
	if (s == NULL || String_init(&ost, strlen(s))) {
		return NULL;
	}
	for (c = (unsigned char *)s; *c != '\0'; c ++) {
		if (*c <= 0x80) {
			if (seq != ASCII) {
				if (String_append(&ost, "\x1b\x28\x42", 3)) {
					return NULL;
				}
				seq = ASCII;
			}
			if (String_append_char(&ost, *c)) {
				return NULL;
			}
		} else if (*c > 0x9f && *c < 0xe0) {
			const char table[][3] = {
				"\x21\x21", "\x21\x23", "\x21\x56", "\x21\x57", "\x21\x22",
				"\x21\x26", "\x25\x72", "\x25\x21", "\x25\x23", "\x25\x25",
				"\x25\x27", "\x25\x29", "\x25\x63", "\x25\x65", "\x25\x67",
				"\x25\x43", "\x21\x3c", "\x25\x22", "\x25\x24", "\x25\x26",
				"\x25\x28", "\x25\x2a", "\x25\x2b", "\x25\x2d", "\x25\x2f",
				"\x25\x31", "\x25\x33", "\x25\x35", "\x25\x37", "\x25\x39",
				"\x25\x3b", "\x25\x3d", "\x25\x3f", "\x25\x41", "\x25\x44",
				"\x25\x46", "\x25\x48", "\x25\x4a", "\x25\x4b", "\x25\x4c",
				"\x25\x4d", "\x25\x4e", "\x25\x4f", "\x25\x52", "\x25\x55",
				"\x25\x58", "\x25\x5b", "\x25\x5e", "\x25\x5f", "\x25\x60",
				"\x25\x61", "\x25\x62", "\x25\x64", "\x25\x66", "\x25\x68",
				"\x25\x69", "\x25\x6a", "\x25\x6b", "\x25\x6c", "\x25\x6d",
				"\x25\x6f", "\x25\x73", "\x21\x2b", "\x21\x2c",
			};
			if (seq != KANJI) {
				if (String_append(&ost, "\x1b\x24\x42", 3)) {
					return NULL;
				}
				seq = KANJI;
			}
			if (String_append(&ost, table[*c - 0xa0], 2)) {
				return NULL;
			}
		} else {
			if (seq != KANJI) {
				if (String_append(&ost, "\x1b\x24\x42", 3)) {
					return NULL;
				}
				seq = KANJI;
			}
			if (*c < 0xa0) {
				h = (*c - 0x81) << 1;
			} else {
				h = (*c - 0xc1) << 1;
			}
			c ++;
			if (*c == '\0') {
				if (String_append(&ost, "\x1b\x28\x42", 3)) {
					return NULL;
				}
				break;
			}
			if (*c >= 0x9f ) {
				if (String_append_char(&ost, h + 0x22)) {
					return NULL;
				}
				if (String_append_char(&ost, *c - 0x7e)) {
					return NULL;
				}
			} else if (*c < 0x7f) {
				if (String_append_char(&ost, h + 0x21)) {
					return NULL;
				}
				if (String_append_char(&ost, *c - 0x1f)) {
					return NULL;
				}
			} else {
				if (String_append_char(&ost, h + 0x21)) {
					return NULL;
				}
				if (String_append_char(&ost, *c - 0x20)) {
					return NULL;
				}
			}
		}
	}
	if (seq != ASCII) {
		if (String_append(&ost, "\x1b\x28\x42", 3)) {
			return NULL;
		}
	}
	return ost.str;
}

char *sjis2euc(char *s)
{
	String ost;
	unsigned char *c, h;
	if (s == NULL || String_init(&ost, strlen(s))) {
		return NULL;
	}
	for (c = (unsigned char *)s; *c != '\0'; c ++) {
		if (*c <= 0x80) {
			if (String_append_char(&ost, *c)) {
				return NULL;
			}
		} else if (*c > 0x9f && *c < 0xe0) {
			if (String_append_char(&ost, '\x8e')) {
				return NULL;
			}
			if (String_append_char(&ost, (*c - 0x80) | 0x80)) {
				return NULL;
			}
		} else {
			if (*c < 0xa0) {
				h = (*c - 0x81) << 1;
			} else {
				h = (*c - 0xc1) << 1;
			}
			c ++;
			if (*c == '\0') {
				break;
			}
			if (*c >= 0x9f ) {
				if (String_append_char(&ost, (h + 0x22) | 0x80)) {
					return NULL;
				}
				if (String_append_char(&ost, (*c - 0x7e) | 0x80)) {
					return NULL;
				}
			} else if (*c < 0x7f) {
				if (String_append_char(&ost, (h + 0x21) | 0x80)) {
					return NULL;
				}
				if (String_append_char(&ost, (*c - 0x1f) | 0x80)) {
					return NULL;
				}
			} else {
				if (String_append_char(&ost, (h + 0x21) | 0x80)) {
					return NULL;
				}
				if (String_append_char(&ost, (*c - 0x20) | 0x80)) {
					return NULL;
				}
			}
		}
	}
	return ost.str;
}

char *han2zen(char *s)
{
	String ost;
	unsigned char *c;
	if (s == NULL || String_init(&ost, strlen(s))) {
		return NULL;
	}
	for (c = (unsigned char *)s; *c != '\0'; c ++) {
		if (*c <= 0x80) {
			if (String_append_char(&ost, *c)) {
				return NULL;
			}
		} else if (*c > 0x9f && *c < 0xe0) {
			const char table[][3] = {
				"\x81\x40", "\x81\x42", "\x81\x75", "\x81\x76", "\x81\x41",
				"\x81\x45", "\x83\x92", "\x83\x40", "\x83\x42", "\x83\x44",
				"\x83\x46", "\x83\x48", "\x83\x83", "\x83\x85", "\x83\x87",
				"\x83\x62", "\x81\x5b", "\x83\x41", "\x83\x43", "\x83\x45",
				"\x83\x47", "\x83\x49", "\x83\x4a", "\x83\x4c", "\x83\x4e",
				"\x83\x50", "\x83\x52", "\x83\x54", "\x83\x56", "\x83\x58",
				"\x83\x5a", "\x83\x5c", "\x83\x5e", "\x83\x60", "\x83\x63",
				"\x83\x65", "\x83\x67", "\x83\x69", "\x83\x6a", "\x83\x6b",
				"\x83\x6c", "\x83\x6d", "\x83\x6e", "\x83\x71", "\x83\x74",
				"\x83\x77", "\x83\x7a", "\x83\x7d", "\x83\x7e", "\x83\x80",
				"\x83\x81", "\x83\x82", "\x83\x84", "\x83\x86", "\x83\x88",
				"\x83\x89", "\x83\x8a", "\x83\x8b", "\x83\x8c", "\x83\x8d",
				"\x83\x8f", "\x83\x93", "\x81\x4a", "\x81\x4b", 
			};
			if (String_append(&ost, table[*c - 0xa0], 2)) {
				return NULL;
			}
		} else {
			if (String_append_char(&ost, *c)) {
				return NULL;
			}
			c ++;
			if (*c == '\0') {
				break;
			}
			if (String_append_char(&ost, *c)) {
				return NULL;
			}
		}
	}
	return ost.str;
}

char *escape(char *s)
{
	String ost;
	if (s == NULL || String_init(&ost, strlen(s))) {
		return NULL;
	}
	for (; *s != '\0'; s ++) {
		if ((*s >= 'a' && *s <= 'z') || (*s >= 'A' && *s <= 'Z') ||
			(*s >= '0' && *s <= '9') || *s == '_' || *s == '.' || *s == '-') {
			if (String_append_char(&ost, *s)) {
				return NULL;
			}
		} else if (*s == ' ') {
			if (String_append_char(&ost, '+')) {
				return NULL;
			}
		} else {
			char tmp[4];
			sprintf(tmp, "%%%02X", (unsigned char)*s);
			if (String_append(&ost, tmp, 3)) {
				return NULL;
			}
		}
	}
	return ost.str;
}

#ifndef KMELEON_PLUGIN_EXPORTS

int main(int argc, char *argv[])
{
	char buf[1024], *p;
	FILE *sp, *jp, *ep, *zp;
	if (argc < 1) {
		return 0;
	}
	sp = fopen(argv[1], "r");
	jp = fopen("jis.txt", "w");
	ep = fopen("euc.txt", "w");
	zp = fopen("zen.txt", "w");
	while (fgets(buf, sizeof(buf), sp) != NULL) {
		p = sjis2jis(buf);
		if (p) {
			fputs(p, jp);
			free(p);
		}
		p = sjis2euc(buf);
		if (p) {
			fputs(p, ep);
			free(p);
		}
		p = han2zen(buf);
		if (p) {
			fputs(p, zp);
			free(p);
		}
		p = escape(buf);
		if (p) {
			fputs(p, stderr);
			free(p);
		}
	}
	fclose(sp);
	fclose(jp);
	fclose(ep);
	return 0;
}

#endif
